/*
 *  Copyright 2013-2050 Emmanuel BRUN (contact@amapj.fr)
 * 
 *  This file is part of AmapJ.
 *  
 *  AmapJ is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Affero General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.

 *  AmapJ is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Affero General Public License for more details.
 *
 *  You should have received a copy of the GNU Affero General Public License
 *  along with AmapJ.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 */
 package fr.amapj.service.engine.objectstorage;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import javax.persistence.EntityManager;

import org.apache.commons.io.IOUtils;

import fr.amapj.common.AmapjRuntimeException;
import fr.amapj.view.engine.ui.AppConfiguration;
import io.minio.GetObjectArgs;
import io.minio.ListObjectsArgs;
import io.minio.MinioClient;
import io.minio.PutObjectArgs;
import io.minio.RemoveObjectArgs;
import io.minio.Result;
import io.minio.messages.Item;


/**
 * Exemple d'utilisation de minio : https://github.com/minio/minio-java/blob/release/examples/GetObject.java 
 * 
 *
 */
public class RemoteObjectStorageService implements ObjectStorageService
{
	private String s3StorageBucket = null;
	private MinioClient minioClient;
		
	RemoteObjectStorageService(AppConfiguration app)
	{
		s3StorageBucket = app.getS3StorageBucket();
		
		minioClient = MinioClient.builder()
			              .endpoint(app.getS3StorageEndPoint())
			              .credentials(app.getS3StorageLogin(), app.getS3StoragePassword())
			              .build();
	}
	
	/** 
	 * content doit être non null
	 */
	public void updateOrCreateObject(EntityManager em,String key,byte[] content,boolean compress)
	{
	    // 
        ByteArrayInputStream bais;
        if (compress)
        {
        	try
    		{
	        	ByteArrayOutputStream baos = new ByteArrayOutputStream();
				GZIPOutputStream os = new GZIPOutputStream(baos);
				os.write(content);
				os.close();
				bais = new ByteArrayInputStream(baos.toByteArray());
    		} 
    		catch (IOException e)
    		{
    			throw new AmapjRuntimeException(e);
    		}
        }
        else
        {
        	bais = new ByteArrayInputStream(content);
        }

        // 
        try 
        {
			minioClient.putObject(PutObjectArgs.builder().bucket(s3StorageBucket).object(key).stream(bais, bais.available(), -1).build());
			bais.close();
		} 
        catch (Exception e) 
        {
			throw new AmapjRuntimeException(e);
		}
	}
		
	
	/**
	 * Retourne null si l'objet n'existe pas 
	 */
	public byte[] loadObject(EntityManager em,String key,boolean uncompress)
	{
		try
		{
			InputStream stream = minioClient.getObject(GetObjectArgs.builder().bucket(s3StorageBucket).object(key).build());
			if (uncompress)
			{
				stream = new GZIPInputStream(stream);
			}
			byte[] res = IOUtils.toByteArray(stream);
			stream.close();
			return res;
		}
		catch (Exception e) 
        {
			throw new AmapjRuntimeException(e);
		}
	}

	public void deleteObject(EntityManager em, String key)
	{
		try
		{
			minioClient.removeObject(RemoveObjectArgs.builder().bucket(s3StorageBucket).object(key).build());	
		} 
	    catch (Exception e) 
	    {
			throw new AmapjRuntimeException(e);
		}
	}
	
	public static void main(String[] args) throws Exception
	{
		
		MinioClient m = MinioClient.builder()
	              .endpoint("http://127.0.0.1:9000/")
	              .credentials("s3login", "s3password")
	              .build();
		
		/*
		// Get information of an object.
        ObjectStat objectStat = m.statObject(StatObjectArgs.builder().bucket("abucket").object("amap1/").build());
        System.out.println(objectStat);
        */
		
		// Permet d'obtenir en une seule requete la liste de tous les objets dans le repertoire amap2
		// Ceci ce traduit par une seule requete HTTP 
	    Iterable<Result<Item>> results = m.listObjects(ListObjectsArgs.builder()
	                    .bucket("abucket")
	                    .prefix("amap2")
	                    .recursive(true)	               
	                    .build());

	        for (Result<Item> result : results) 
	        {
	          Item item = result.get();
	          System.out.println(/*item.lastModified() +*/ "\t" + item.size() + "\t" + item.objectName());
	        }
	}
	
	
}
